@@ -1,3 +1,19 @@
+RES_DATASET <- c(
+    "id",
+    "mip_era", "activity_drs", "institution_id", "source_id",
+    "experiment_id", "member_id", "table_id", "frequency", "grid_label",
+    "version", "nominal_resolution", "variable_id", "variable_long_name",
+    "variable_units", "data_node", "pid"
+)
+
+RES_FILE <- c(
+    "id", "dataset_id",
+    "mip_era", "activity_drs", "institution_id", "source_id",
+    "experiment_id", "member_id", "table_id", "frequency", "grid_label",
+    "version", "nominal_resolution", "variable_id", "variable_long_name",
+    "variable_units", "data_node", "size", "url", "tracking_id"
+)
+
 # esgf_query {{{
 #' Query CMIP6 data using ESGF search RESTful API
 #'
@@ -8,11 +24,14 @@
 #'
 #' The ESGF search service exposes a RESTful URL that can be used by clients to
 #' query the contents of the underlying search index, and return results
-#' matching the given constraints. ributed capabilities of the ESGF search, the
-#' URL at any Index Node can be used to query that Node only, or all Nodes in
-#' the ESGF system. `esgf_query()` uses the [LLNL (Lawrence Livermore National
-#' Laboratory)](http://esgf-node.llnl.gov)
-#' Index Node.
+#' matching the given constraints. With the distributed capabilities of the ESGF
+#' search, the URL at any Index Node can be used to query that Node only, or all
+#' Nodes in the ESGF system. `esgf_query()` uses the
+#' [LLNL (Lawrence Livermore National Laboratory) Index Node](http://esgf-node.llnl.gov).
+#'
+#' The core Controlled Vocabularies (CVs) for use in CMIP6, including all
+#' activities, experiment, sources (GCMs), frequencies can be found at the
+#' [WCRP-CMIP/CMIP6_CVs](https://github.com/WCRP-CMIP/CMIP6_CVs) GitHub repo.
 #'
 #' @param activity A character vector indicating activity identifiers. Default:
 #'        `"ScenarioMIP"`. Possible values:
@@ -56,7 +75,7 @@
 #' * `hursmin`: Minimum near-surface relative humidity, units: `%`.
 #' * `psl`: Sea level pressure, units: `Pa`.
 #' * `rsds`: Surface downwelling shortwave radiation, units: `W m-2`.
-#' * `rlds`: Surface downwelling longware radiation, units: `W m-2`.
+#' * `rlds`: Surface downwelling longwave radiation, units: `W m-2`.
 #' * `sfcWind`: Near-surface (usually, 10 meters) wind speed, units: `m s-1`.
 #' * `pr`: Precipitation, units: `kg m-2 s-1`.
 #' * `clt`: Total cloud area fraction for the whole atmospheric column, as
@@ -101,13 +120,14 @@
 #'        described below. Default: `"r1i1p1f1"`.
 #'        If `NULL`, all possible variants are returned.
 #'
-#' * `r`: realization_index (<k>) = realization number (integer >0)
-#' * `i`: initialization_index (<l>) = index for variant of initialization method (integer >0)
-#' * `p`: physics_index (<m>) = index for model physics variant (integer >0)
-#' * `f`: forcing_index (<n>) = index for variant of forcing (integer >0)
+#' * `r`: realization_index (`<k>`) = realization number (integer >0)
+#' * `i`: initialization_index (`<l>`) = index for variant of initialization method (integer >0)
+#' * `p`: physics_index (`<m>`) = index for model physics variant (integer >0)
+#' * `f`: forcing_index (`<n>`) = index for variant of forcing (integer >0)
 #'
 #' @param replica Whether the record is the "master" copy, or a replica. Use
 #'        `FALSE` to return only originals and `TRUE` to return only replicas.
+#'        Use `NULL` to return both the master and the replicas.
 #'        Default: `FALSE`.
 #'
 #' @param latest Whether the record is the latest available version, or a
@@ -195,23 +215,22 @@
 #' @importFrom checkmate assert_character assert_choice assert_count assert_flag assert_subset
 #' @importFrom data.table data.table setattr
 #' @export
-esgf_query <- function (
-    activity = "ScenarioMIP",
-    variable = c("tas", "tasmax", "tasmin", "hurs", "hursmax", "hursmin", "pr", "rsds", "rlds", "psl", "sfcWind", "clt"),
-    frequency = "day",
-    experiment = c("ssp126", "ssp245", "ssp370", "ssp585"),
-    source = c("AWI-CM-1-1-MR", "BCC-CSM2-MR", "CESM2", "CESM2-WACCM",
-               "EC-Earth3", "EC-Earth3-Veg", "GFDL-ESM4", "INM-CM4-8",
-               "INM-CM5-0", "MPI-ESM1-2-HR", "MRI-ESM2-0"),
-    variant = "r1i1p1f1",
-    replica = FALSE,
-    latest = TRUE,
-    resolution = c("100 km", "50 km"),
-    type = "Dataset",
-    limit = 10000L,
-    data_node = NULL
-)
-{
+esgf_query <- function(activity = "ScenarioMIP",
+                       variable = c("tas", "tasmax", "tasmin", "hurs", "hursmax", "hursmin", "pr", "rsds", "rlds", "psl", "sfcWind", "clt"),
+                       frequency = "day",
+                       experiment = c("ssp126", "ssp245", "ssp370", "ssp585"),
+                       source = c(
+                           "AWI-CM-1-1-MR", "BCC-CSM2-MR", "CESM2", "CESM2-WACCM",
+                           "EC-Earth3", "EC-Earth3-Veg", "GFDL-ESM4", "INM-CM4-8",
+                           "INM-CM5-0", "MPI-ESM1-2-HR", "MRI-ESM2-0"
+                       ),
+                       variant = "r1i1p1f1",
+                       replica = FALSE,
+                       latest = TRUE,
+                       resolution = c("100 km", "50 km"),
+                       type = "Dataset",
+                       limit = 10000L,
+                       data_node = NULL) {
     assert_subset(activity, empty.ok = FALSE, choices = c(
         "AerChemMIP", "C4MIP", "CDRMIP", "CFMIP", "CMIP", "CORDEX", "DAMIP",
         "DCPP", "DynVarMIP", "FAFMIP", "GMMIP", "GeoMIP", "HighResMIP",
@@ -227,63 +246,95 @@ esgf_query <- function (
     assert_character(source, any.missing = FALSE, null.ok = TRUE)
     assert_character(variant, any.missing = FALSE, pattern = "r\\d+i\\d+p\\d+f\\d+", null.ok = TRUE)
     assert_character(resolution, any.missing = FALSE, null.ok = TRUE)
-    assert_flag(replica)
+    assert_flag(replica, null.ok = TRUE)
     assert_flag(latest)
     assert_count(limit, positive = TRUE)
     assert_choice(type, choices = c("Dataset", "File"))
     assert_character(data_node, any.missing = FALSE, null.ok = TRUE)
 
-    url_base <- "http://esgf-node.llnl.gov/esg-search/search/?"
+    url_base <- "https://esgf-node.llnl.gov/esg-search/search?"
 
-    dict <- c(activity = "activity_id",
-              experiment = "experiment_id",
-              source = "source_id",
-              variable = "variable_id",
-              resolution = "nominal_resolution",
-              variant = "variant_label")
+    dict <- c(
+        activity = "activity_id",
+        experiment = "experiment_id",
+        source = "source_id",
+        variable = "variable_id",
+        resolution = "nominal_resolution",
+        variant = "variant_label"
+    )
+
+    pair <- function(x, encode = TRUE) {
+        checkmate::assert_vector(x, TRUE, null.ok = TRUE)
 
-    pair <- function (x, first = FALSE) {
         # get name
         var <- deparse(substitute(x))
+
         # skip if empty
-        if (is.null(x) || length(x) == 0) return()
+        if (is.null(x) || length(x) == 0) {
+            return()
+        }
         # get key name
         key <- dict[names(dict) == var]
         if (!length(key)) key <- var
+
         if (is.logical(x)) x <- tolower(x)
-        s <- paste0(key, "=", paste0(x, collapse = "%2C")) # %2C = ","
-        if (first) s else paste0("&", s)
+
+        if (encode) x <- query_param_encode(as.character(x))
+
+        paste0(key, "=", paste0(x, collapse = query_param_encode(",")))
     }
 
-    `%and%` <- function (lhs, rhs) if (is.null(rhs)) lhs else paste0(lhs, rhs)
+    `%and%` <- function(lhs, rhs) {
+        if (is.null(rhs)) {
+            lhs
+        } else if (lhs == url_base) {
+            paste(lhs, rhs, sep = "", collapse = "")
+        } else {
+            paste(lhs, rhs, sep = "&", collapse = "&")
+        }
+    }
 
     project <- "CMIP6"
-    format <- "application%2Fsolr%2Bjson"
+    format <- "application/solr+json"
+    offset <- 0L
 
-    resolution <- c(gsub(" ", "", resolution, fixed = TRUE),
-                    gsub(" ", "+", resolution, fixed = TRUE))
+    resolution <- c(
+        gsub(" ", "", resolution, fixed = TRUE),
+        gsub(" ", "+", resolution, fixed = TRUE)
+    )
+
+    # use `fileds` to directly subset data from responses
+    if (type == "Dataset") {
+        fields <- RES_DATASET
+    } else if (type == "File") {
+        fields <- RES_FILE
+    }
 
     q <- url_base %and%
-        pair(project, TRUE) %and%
+        pair(offset) %and%
+        pair(limit) %and%
+        pair(type) %and%
+        pair(replica) %and%
+        pair(latest) %and%
+        pair(project) %and%
         pair(activity) %and%
         pair(experiment) %and%
         pair(source) %and%
         pair(variable) %and%
-        pair(resolution) %and%
+        pair(resolution, FALSE) %and%
         pair(variant) %and%
         pair(data_node) %and%
         pair(frequency) %and%
-        pair(replica) %and%
-        pair(latest) %and%
-        pair(type) %and%
-        pair(limit) %and%
+        pair(fields) %and%
         pair(format)
 
-    q <- tryCatch(jsonlite::read_json(q), warning = function (w) w, error = function (e) e)
+    q <- tryCatch(jsonlite::read_json(q), warning = function(w) w, error = function(e) e)
 
+    # nocov start
     if (inherits(q, "warning") || inherits(q, "error")) {
         message("No matched data. Please check network connection and the availability of LLNL ESGF node.")
         dt <- data.table::data.table()
+        # nocov end
     } else if (q$response$numFound == 0L) {
         message("No matched data. Please examine the actual response using 'attr(x, \"response\")'.")
         dt <- data.table::data.table()
@@ -301,42 +352,39 @@ esgf_query <- function (
 
 # extract_query_dataset {{{
 #' @importFrom data.table rbindlist
-extract_query_dataset <- function (q) {
-    dt <- data.table::rbindlist(lapply(q$response$docs, function (l) {
-        l <- l[c("id", "mip_era", "activity_drs", "institution_id", "source_id",
-            "experiment_id", "member_id", "table_id", "frequency", "grid_label",
-            "version", "nominal_resolution", "variable_id", "variable_long_name",
-            "variable_units", "data_node", "pid")]
-        lapply(l, unlist)
-    }))
+extract_query_dataset <- function(q) {
+    dt <- data.table::rbindlist(lapply(q$response$docs, lapply, unlist))
+    data.table::set(dt, NULL, setdiff(names(dt), RES_DATASET), NULL)
+    data.table::setcolorder(dt, RES_DATASET)
     data.table::setnames(dt, c("id", "pid"), c("dataset_id", "dataset_pid"))
 }
 # }}}
 
 # extract_query_file {{{
 #' @importFrom data.table rbindlist setcolorder setnames tstrsplit ":="
-extract_query_file <- function (q) {
+extract_query_file <- function(q) {
     # to avoid No visible binding for global variable check NOTE
     id <- NULL
-    dt_file <- data.table::rbindlist(lapply(q$response$docs, function (l) {
-        l <- l[c("id", "dataset_id", "mip_era", "activity_drs", "institution_id",
-            "source_id", "experiment_id", "member_id", "table_id", "frequency",
-            "grid_label", "version", "nominal_resolution", "variable_id",
-            "variable_long_name", "variable_units", "data_node", "size", "url",
-            "tracking_id")]
+    dt <- data.table::rbindlist(lapply(q$response$docs, function(l) {
         l$url <- grep("HTTPServer", unlist(l$url), fixed = TRUE, value = TRUE)
+        # nocov start
         if (!length(l$url)) {
             warning("Dataset with id '", l$id, "' does not have a HTTPServer download method.")
             l$url <- NA_character_
         }
+        # nocov end
         lapply(l, unlist)
     }))
+    data.table::set(dt, NULL, setdiff(names(dt), RES_FILE), NULL)
+    data.table::setcolorder(dt, RES_FILE)
 
-    dt_file[, c("datetime_start", "datetime_end") := parse_file_date(id, frequency)]
-    dt_file[, url := gsub("\\|.+$", "", url)]
-    data.table::setnames(dt_file, c("id", "size", "url"),
-        c("file_id", "file_size", "file_url"))
-    data.table::setcolorder(dt_file, c(
+    dt[, c("datetime_start", "datetime_end") := parse_file_date(id, frequency)]
+    dt[, url := gsub("\\|.+$", "", url)]
+    data.table::setnames(
+        dt, c("id", "size", "url"),
+        c("file_id", "file_size", "file_url")
+    )
+    data.table::setcolorder(dt, c(
         "file_id", "dataset_id", "mip_era", "activity_drs", "institution_id",
         "source_id", "experiment_id", "member_id", "table_id", "frequency",
         "grid_label", "version", "nominal_resolution", "variable_id",
@@ -344,7 +392,7 @@ extract_query_file <- function (q) {
         "datetime_end", "file_size", "data_node", "file_url", "tracking_id"
     ))
 
-    dt_file
+    dt
 }
 # }}}
 
@@ -404,43 +452,49 @@ extract_query_file <- function (q) {
 #' @importFrom data.table copy fwrite rbindlist set setcolorder
 #' @importFrom rappdirs user_data_dir
 #' @export
-init_cmip6_index <- function (
-    activity = "ScenarioMIP",
-    variable = c("tas", "tasmax", "tasmin", "hurs", "hursmax", "hursmin", "pr", "rsds", "rlds", "psl", "sfcWind", "clt"),
-    frequency = "day",
-    experiment = c("ssp126", "ssp245", "ssp370", "ssp585"),
-    source = c("AWI-CM-1-1-MR", "BCC-CSM2-MR", "CESM2", "CESM2-WACCM",
-               "EC-Earth3", "EC-Earth3-Veg", "GFDL-ESM4", "INM-CM4-8",
-               "INM-CM5-0", "MPI-ESM1-2-HR", "MRI-ESM2-0"),
-    variant = "r1i1p1f1",
-    replica = FALSE,
-    latest = TRUE,
-    resolution = c("100 km", "50 km"),
-    limit = 10000L,
-    data_node = NULL,
-    years = NULL,
-    save = FALSE
-)
-{
+init_cmip6_index <- function(activity = "ScenarioMIP",
+                             variable = c("tas", "tasmax", "tasmin", "hurs", "hursmax", "hursmin", "pr", "rsds", "rlds", "psl", "sfcWind", "clt"),
+                             frequency = "day",
+                             experiment = c("ssp126", "ssp245", "ssp370", "ssp585"),
+                             source = c(
+                                 "AWI-CM-1-1-MR", "BCC-CSM2-MR", "CESM2", "CESM2-WACCM",
+                                 "EC-Earth3", "EC-Earth3-Veg", "GFDL-ESM4", "INM-CM4-8",
+                                 "INM-CM5-0", "MPI-ESM1-2-HR", "MRI-ESM2-0"
+                             ),
+                             variant = "r1i1p1f1",
+                             replica = FALSE,
+                             latest = TRUE,
+                             resolution = c("100 km", "50 km"),
+                             limit = 10000L,
+                             data_node = NULL,
+                             years = NULL,
+                             save = FALSE) {
     assert_integerish(years, lower = 1900, unique = TRUE, sorted = TRUE, any.missing = FALSE, null.ok = TRUE)
     assert_flag(save)
 
     verbose("Querying CMIP6 Dataset Information")
-    qd <- esgf_query(activity = activity, variable = variable, frequency = frequency,
+    qd <- esgf_query(
+        activity = activity, variable = variable, frequency = frequency,
         experiment = experiment, source = source, replica = replica, latest = latest,
         variant = variant, resolution = resolution, limit = limit, type = "Dataset",
-        data_node = data_node)
+        data_node = data_node
+    )
 
-    if (!nrow(qd)) return(qd)
+    if (!nrow(qd)) {
+        return(qd)
+    }
 
+    # nocov start
     # give a warning if dataset query response hits the limits
     if (nrow(qd) == 10000L) {
-        warning("The dataset query returns 10,000 results which ",
+        warning(
+            "The dataset query returns 10,000 results which ",
             "hits the maximum record limitation of a single query using ESGF search RESTful API. ",
             "It is possible that the returned Dataset query responses are not complete. ",
             "It is suggested to examine and refine your query."
         )
     }
+    # nocov end
 
     dt <- data.table::set(qd, NULL, "file_url", NA_character_)
 
@@ -456,8 +510,10 @@ init_cmip6_index <- function (
         .SD <- NULL
 
         # use qd to construction query for files
-        q <- unique(nf[, .SD, .SDcols = c("activity_drs", "source_id", "member_id",
-            "experiment_id", "nominal_resolution", "table_id", "frequency", "variable_id")])
+        q <- unique(nf[, .SD, .SDcols = c(
+            "activity_drs", "source_id", "member_id",
+            "experiment_id", "nominal_resolution", "table_id", "frequency", "variable_id"
+        )])
 
         qf <- esgf_query(
             activity = unique(q$activity_drs),
@@ -474,12 +530,14 @@ init_cmip6_index <- function (
         )
 
         # remove all common columns in file query except for "dataset_id"
-        data.table::set(qf, NULL, value = NULL,
+        data.table::set(qf, NULL,
+            value = NULL,
             setdiff(intersect(names(qd), names(qf)), c("dataset_id", "file_url"))
         )
 
         # remove all common column in nf except for "dataset_id"
-        data.table::set(nf, NULL, value = NULL,
+        data.table::set(nf, NULL,
+            value = NULL,
             setdiff(intersect(names(qf), names(nf)), c("dataset_id"))
         )
 
@@ -487,11 +545,14 @@ init_cmip6_index <- function (
     }
 
     verbose("Checking if data is complete")
+    # nocov start
     if (anyNA(dt$file_url)) {
-        warning("There are still ", length(unique(dt$dataset_id[is.na(dt$file_url)])), " Dataset that ",
+        warning(
+            "There are still ", length(unique(dt$dataset_id[is.na(dt$file_url)])), " Dataset that ",
             "did not find any matched output file after ", retry, " retries."
         )
     }
+    # nocov end
 
     data.table::setcolorder(dt, c(
         "file_id", "dataset_id", "mip_era", "activity_drs", "institution_id",
@@ -515,7 +576,8 @@ init_cmip6_index <- function (
 
         dt[, `:=`(expect_start = datetime_start, expect_end = datetime_end)]
         dt <- dt[exp, on = c("expect_start<=expect_end", "expect_end>=expect_start")][
-            , `:=`(expect_start = NULL, expect_end = NULL)]
+            , `:=`(expect_start = NULL, expect_end = NULL)
+        ]
     }
 
     # remove duplications
@@ -526,7 +588,7 @@ init_cmip6_index <- function (
         data.table::fwrite(dt, file.path(.data_dir(TRUE), "cmip6_index.csv"))
         verbose("Data file index saved to '", normalizePath(file.path(.data_dir(TRUE), "cmip6_index.csv")), "'")
 
-        EPWSHIFTR_ENV$index_db <- data.table::copy(dt)
+        this$index_db <- data.table::copy(dt)
     }
 
     dt
@@ -548,68 +610,73 @@ init_cmip6_index <- function (
 #' }
 #' @importFrom data.table copy fread
 #' @export
-load_cmip6_index <- function (force = FALSE) {
-    if (is.null(EPWSHIFTR_ENV$index_db)) force <- TRUE
+load_cmip6_index <- function(force = FALSE) {
+    if (is.null(this$index_db)) force <- TRUE
 
     if (!force) {
-        idx <- data.table::copy(EPWSHIFTR_ENV$index_db)
+        idx <- data.table::copy(this$index_db)
     } else {
         f <- normalizePath(file.path(.data_dir(force = FALSE), "cmip6_index.csv"), mustWork = FALSE)
         if (!file.exists(f)) {
             stop(sprintf("CMIP6 experiment output file index does not exists. You may want to create one using 'init_cmip6_index()'."))
         }
 
+        # nocov start
         # load file info
         idx <- tryCatch(
             data.table::fread(f, colClasses = c("version" = "character", "file_size" = "double")),
-            error = function (e) {
-                stop("Failed to parse CMIP6 experiment output file index.\n", conditionMessage(e))
+            warning = function(w) {
+                stop("Failed to parse CMIP6 experiment output file index. ", conditionMessage(w))
+            },
+            error = function(e) {
+                stop("Failed to parse CMIP6 experiment output file index. ", conditionMessage(e))
             }
         )
-        message("Loading CMIP6 experiment output file index created at ", file.info(f)$mtime, ".")
+        # nocov end
+        verbose("Loading CMIP6 experiment output file index created at ", as.character(file.info(f)$mtime), ".")
+    }
 
-        # fix column types in case of empty values
-        if ("file_path" %in% names(idx)) {
-            data.table::set(idx, NULL, "file_path", as.character(idx$file_path))
-            idx[J(""), on = "file_path", file_path := NA_character_]
+    # fix column types in case of empty values
+    if ("file_path" %in% names(idx)) {
+        data.table::set(idx, NULL, "file_path", as.character(idx$file_path))
+        idx[J(""), on = "file_path", file_path := NA_character_]
+    }
+    if ("file_realsize" %in% names(idx)) {
+        data.table::set(idx, NULL, "file_realsize", as.numeric(idx$file_realsize))
+    }
+    if ("file_mtime" %in% names(idx)) {
+        # to avoid No visible binding for global variable check NOTE
+        file_mtime <- NULL
+        if (is.character(idx$file_mtime)) {
+            idx[J(""), on = "file_mtime", file_mtime := NA]
         }
-        if ("file_realsize" %in% names(idx)) {
-            data.table::set(idx, NULL, "file_realsize", as.numeric(idx$file_realsize))
+        idx[, file_mtime := setattr(as.POSIXct(file_mtime, origin = "1970-01-01"), "tzone", NULL)]
+    }
+    if ("time_units" %in% names(idx)) {
+        data.table::set(idx, NULL, "time_units", as.character(idx$time_units))
+    }
+    if ("time_calendar" %in% names(idx)) {
+        data.table::set(idx, NULL, "time_calendar", as.character(idx$time_calendar))
+    }
+    if ("datetime_start" %in% names(idx)) {
+        # to avoid No visible binding for global variable check NOTE
+        datetime_start <- NULL
+        if (is.character(idx$datetime_start)) {
+            idx[J(""), on = "datetime_start", datetime_start := NA]
         }
-        if ("file_mtime" %in% names(idx)) {
-            # to avoid No visible binding for global variable check NOTE
-            file_mtim <- NULL
-            if (is.character(idx$file_mtime)) {
-                idx[J(""), on = "file_mtime", file_mtime := NA]
-            }
-            idx[, file_mtime := as.POSIXct(file_mtime, origin = "1970-01-01", Sys.timezone())]
-        }
-        if ("time_units" %in% names(idx)) {
-            data.table::set(idx, NULL, "time_units", as.character(idx$time_units))
-        }
-        if ("time_calendar" %in% names(idx)) {
-            data.table::set(idx, NULL, "time_calendar", as.character(idx$time_calendar))
-        }
-        if ("datetime_start" %in% names(idx)) {
-            # to avoid No visible binding for global variable check NOTE
-            datetime_start <- NULL
-            if (is.character(idx$datetime_start)) {
-                idx[J(""), on = "datetime_start", datetime_start := NA]
-            }
-            data.table::set(idx, NULL, "datetime_start", as.POSIXct(idx$datetime_start, "UTC"))
-        }
-        if ("datetime_end" %in% names(idx)) {
-            # to avoid No visible binding for global variable check NOTE
-            datetime_end <- NULL
-            if (is.character(idx$datetime_end)) {
-                idx[J(""), on = "datetime_end", datetime_end := NA]
-            }
-            data.table::set(idx, NULL, "datetime_end", as.POSIXct(idx$datetime_end, "UTC"))
+        data.table::set(idx, NULL, "datetime_start", as.POSIXct(idx$datetime_start, "UTC"))
+    }
+    if ("datetime_end" %in% names(idx)) {
+        # to avoid No visible binding for global variable check NOTE
+        datetime_end <- NULL
+        if (is.character(idx$datetime_end)) {
+            idx[J(""), on = "datetime_end", datetime_end := NA]
         }
+        data.table::set(idx, NULL, "datetime_end", as.POSIXct(idx$datetime_end, "UTC"))
     }
 
     # udpate package internal stored file index
-    EPWSHIFTR_ENV$index_db <- data.table::copy(idx)
+    this$index_db <- data.table::copy(idx)
 
     idx[]
 }
@@ -633,16 +700,19 @@ load_cmip6_index <- function (force = FALSE) {
 #' @return A [data.table::data.table()].
 #' @importFrom checkmate assert_data_table
 #' @export
-set_cmip6_index <- function (index, save = FALSE) {
+set_cmip6_index <- function(index, save = FALSE) {
     checkmate::assert_data_table(index)
-    checkmate::assert_subset(names(index),
-        c("file_id", "dataset_id", "mip_era", "activity_drs", "institution_id",
-        "source_id", "experiment_id", "member_id", "table_id", "frequency",
-        "grid_label", "version", "nominal_resolution", "variable_id",
-        "variable_long_name", "variable_units", "datetime_start",
-        "datetime_end", "file_size", "data_node", "file_url", "dataset_pid",
-        "tracking_id", "file_path", "file_realsize", "file_mtime", "time_units",
-        "time_calendar")
+    checkmate::assert_subset(
+        names(index),
+        c(
+            "file_id", "dataset_id", "mip_era", "activity_drs", "institution_id",
+            "source_id", "experiment_id", "member_id", "table_id", "frequency",
+            "grid_label", "version", "nominal_resolution", "variable_id",
+            "variable_long_name", "variable_units", "datetime_start",
+            "datetime_end", "file_size", "data_node", "file_url", "dataset_pid",
+            "tracking_id", "file_path", "file_realsize", "file_mtime", "time_units",
+            "time_calendar"
+        )
     )
 
     # save database into the app data directory
@@ -652,7 +722,7 @@ set_cmip6_index <- function (index, save = FALSE) {
     }
 
     # udpate package internal stored file index
-    EPWSHIFTR_ENV$index_db <- data.table::copy(index)
+    this$index_db <- data.table::copy(index)
 
     invisible(index)
 }
@@ -671,7 +741,7 @@ set_cmip6_index <- function (index, save = FALSE) {
 #' get_data_dir()
 #'
 #' @export
-get_data_dir <- function () {
+get_data_dir <- function() {
     .data_dir(force = TRUE)
 }
 # }}}
@@ -695,61 +765,59 @@ get_data_dir <- function () {
 #' | `ping`      | double    | Data node response in milliseconds during speed test                            |
 #'
 #' @examples
-#' \donttest{
+#' \dontrun{
 #' get_data_node()
 #' }
 #'
 #' @export
-get_data_node <- function (speed_test = FALSE, timeout = 3) {
-    # read html page
-    f <- tempfile()
-    utils::download.file("https://esgf-node.llnl.gov/status/", f, "libcurl", quiet = TRUE)
-    l <- readLines(f)
+get_data_node <- function(speed_test = FALSE, timeout = 3) {
+    # use the metagrid-backend to get the data node status
+    # see: https://github.com/esgf2-us/metagrid/blob/2e90dd10317506a82f120217e39c4a3cde6a7560/backend/.envs/.django#L30
+    #      https://github.com/ESGF/esgf-utils/blob/master/node_status/query_prom.py
+    res <- tryCatch(
+        jsonlite::fromJSON("https://aims2.llnl.gov/metagrid-backend/proxy/status"),
+        error = function(e) NULL
+    )
 
-    # locate table
-    l_s <- grep("<!--load block main-->", l, fixed = TRUE)
-    if (!length(l_s)) stop("Internal Error: Failed to read data node table")
-    l <- l[l_s:length(l)]
-    l_s <- grep("<table>", l, fixed = TRUE)[1L]
-    l_e <- grep("</table>", l, fixed = TRUE)[1L]
-    if (!length(l_s) || !length(l_e)) stop("Internal Error: Failed to read data node table")
-    l <- l[l_s:l_e]
+    # nocov start
+    if (is.null(res) || res$status != "success") {
+        message("Failed to retrieve the data node status from aims2.llnl.gov.")
+        return(data.table::data.table())
+    }
+    # nocov end
 
-    # extract nodes
-    loc <- regexec("\\t<td>(.+)</td>", l)
-    nodes <- vapply(seq_along(l), function (i) {
-        if (all(loc[[i]][1] == -1L)) return(NA_character_)
-        substr(l[i], loc[[i]][2], loc[[i]][2] + attr(loc[[i]], "match.length")[2] - 1L)
-    }, NA_character_)
-    nodes <- nodes[!is.na(nodes)]
+    res <- data.table::data.table(
+        data_node = res$data$result$metric$instance,
+        status = data.table::fifelse(
+            vapply(res$data$result$value, .subset2, character(1L), 2L) == "1",
+            "UP", "DOWN"
+        )
+    )
 
-    # extract status
-    loc <- regexec('\\t\\t<font color="#\\S{6}"><b>(UP|DOWN)</b>', l)
-    status <- vapply(seq_along(l), function (i) {
-        if (all(loc[[i]][1] == -1L)) return(NA_character_)
-        substr(l[i], loc[[i]][2], loc[[i]][2] + attr(loc[[i]], "match.length")[2] - 1L)
-    }, NA_character_)
-    status <- status[!is.na(status)]
-
-    if (length(nodes) != length(status)) stop("Internal Error: Failed to read data node table")
-    res <- data.table::data.table(data_node = nodes, status = status)
     data.table::setorderv(res, "status", -1)
 
-    if (!speed_test) return(res)
+    if (!speed_test) {
+        return(res)
+    }
 
+    # nocov start
     if (!requireNamespace("pingr", quietly = TRUE)) {
-        stop("'epwshiftr' relies on the package 'pingr' to perform speed test",
+        stop(
+            "'epwshiftr' relies on the package 'pingr' to perform speed test",
             "please add this to your library with install.packages('pingr') and try again."
         )
     }
+    # nocov end
 
+    # nocov start
     if (!length(nodes_up <- res[status == "UP", data_node])) {
         message("No working data nodes available now. Skip speed test")
         return(res)
     }
+    # nocov end
 
     # use the pingr package to test the connection speed
-    speed <- vapply(nodes_up, function (node) {
+    speed <- vapply(nodes_up, function(node) {
         message(sprintf("Testing data node '%s'...", node))
         pingr::ping(node, count = 1, timeout = timeout)
     }, numeric(1))
@@ -760,32 +828,54 @@ get_data_node <- function (speed_test = FALSE, timeout = 3) {
 
 # parse_file_date {{{
 #' @importFrom data.table fifelse
-parse_file_date <- function (id, frequency) {
-    dig <- fifelse(grepl("hr", frequency, fixed = TRUE), 12L,
-        fifelse(frequency == "day", 8L,
-            fifelse(frequency == "dec", 4L,
-                fifelse(grepl("mon", frequency, fixed = TRUE), 6L,
-                    fifelse(grepl("yr", frequency, fixed = TRUE), 4L, 0L)))))
+parse_file_date <- function(id, frequency) {
+    dig <- fifelse(
+        grepl("hr", frequency, fixed = TRUE), 12L,
+        fifelse(
+            frequency == "day", 8L,
+            fifelse(
+                frequency == "dec", 4L,
+                fifelse(
+                    grepl("mon", frequency, fixed = TRUE), 6L,
+                    fifelse(grepl("yr", frequency, fixed = TRUE), 4L, 0L)
+                )
+            )
+        )
+    )
 
     reg <- sprintf("([0-9]{%i})-([0-9]{%i})", dig, dig)
     reg[dig == 0L] <- NA_character_
 
-    suf <- fifelse(dig == 0L, "",
-        fifelse(dig == 4L, "0101",
-            fifelse(dig == 6L, "01", "")))
+    suf <- fifelse(
+        dig == 0L, "",
+        fifelse(
+            dig == 4L, "0101",
+            fifelse(dig == 6L, "01", "")
+        )
+    )
 
-    fmt <- fifelse(dig == 0L, NA_character_,
-        fifelse(dig == 4L, "%Y%m%d",
-            fifelse(dig == 6L, "%Y%m%d",
-                fifelse(dig == 8L, "%Y%m%d",
-                    fifelse(dig == 12L, "%Y%m%d%H%M%s", NA_character_)))))
+    fmt <- fifelse(
+        dig == 0L, NA_character_,
+        fifelse(
+            dig == 4L, "%Y%m%d",
+            fifelse(
+                dig == 6L, "%Y%m%d",
+                fifelse(
+                    dig == 8L, "%Y%m%d",
+                    fifelse(dig == 12L, "%Y%m%d%H%M%s", NA_character_)
+                )
+            )
+        )
+    )
 
     data.table(id, reg, suf, fmt)[
-        !J(NA_character_), on = "reg", by = "reg",
+        !J(NA_character_),
+        on = "reg", by = "reg",
         c("datetime_start", "datetime_end") := {
-        m <- regexpr(.BY$reg, id)
-        s <- data.table::tstrsplit(regmatches(id, m), "-", fixed = TRUE)
-        lapply(s, function (x) as.POSIXct(paste0(x, suf), format = fmt[1L], tz = "UTC"))
-    }][, .SD, .SDcols = c("datetime_start", "datetime_end")]
+            m <- regexpr(.BY$reg, id)
+            s <- data.table::tstrsplit(regmatches(id, m), "-", fixed = TRUE)
+            lapply(s, function(x) as.POSIXct(paste0(x, suf), format = fmt[1L], tz = "UTC"))
+        }
+    ][, .SD, .SDcols = c("datetime_start", "datetime_end")]
 }
 # }}}
