@@ -1,7 +1,7 @@
 # preprocess_morphing {{{
 #' @importFrom checkmate assert_count
 #' @importFrom data.table set
-preprocess_morphing <- function (dt, leapyear = FALSE, years = NULL, labels = NULL) {
+preprocess_morphing <- function (dt, leapyear = FALSE, years = NULL, labels = NULL, warning = FALSE) {
     # add datetime components
     data.table::set(dt, NULL, c("year", "month", "day", "hour", "minute"),
         list(data.table::year(dt$datetime),
@@ -40,7 +40,35 @@ preprocess_morphing <- function (dt, leapyear = FALSE, years = NULL, labels = NU
         }
     }
 
-    # calculate monthly mean and average value for longitude and latitude
+    if (warning) {
+        # give a warning if the dataset only includes years less than a decade
+        rng_year <- dt[, list(years = list(sort(unique(year))), num_years = length(unique(year))),
+            by = c("variable", "table_id", "source_id", "experiment_id", "member_id")
+        ][num_years < 10L]
+
+        if (nrow(rng_year)) {
+            set(rng_year, NULL, "index_case", seq.int(nrow(rng_year)))
+
+            mes <- rng_year[, by = "index_case", {
+                head <- sprintf("#%i | For case %s_%s_%s_%s_%s:\n",
+                    .BY$index_case, variable, table_id, source_id, experiment_id, member_id
+                )
+                yrs <- sprintf("   --> [%i] %s", num_years, paste0("'", years[[1L]], "'", collapse = ", "))
+                list(message = paste0(head, yrs, collapse = "\n"))
+            }]$message
+
+            ori <- getOption("warning.length")
+            options(warning.length = 8170L)
+            on.exit(options(warning.length = ori), add = TRUE)
+
+            warning("Case(s) shown below contains CMIP data less than a decade. ",
+                "The morphed data may not be able to capture average weather of the future climate.\n",
+                paste0(mes, collapse = "\n"), call. = FALSE)
+        }
+    }
+
+    # calculate monthly mean and average value for longitude, latitude and
+    # distance
     res <- location_mean(dt, c("datetime", "year", "day", "hour", "minute", "second"))
     data.table::set(res, NULL, "value", units::set_units(res$value, res$units[1], mode = "standard"))
 
@@ -101,6 +129,55 @@ remove_units <- function (data, var) {
 #'
 #' The EPW weather variables that get morphed are listed in details.
 #'
+#' # The Morphing procedure
+#'
+#' Here *Morphing* is an algorithm proposed by Belcher etc. 2005
+#' used to morph the present-day observed weather files (here the EPWs)  to
+#' produce future climate weather files. The EPW data is used as the
+#' *'baseline climate'*.
+#'
+#' The first step before morphing is to calculate the monthly means of
+#' climatological variables in the EPW file, denoted by \eqn{<x_0>_m}.
+#' The subscript '0' is to denote the present day weather record, and 'm' is to
+#' denote the month.
+#'
+#' The morphing involves three generic operations, i.e. 1) a shift; 2) a linear
+#' stretch (scaling factor); and 3) a shift and a stretch:
+#'
+#' \deqn{
+#' \textrm{Shift: } x = x_0 + \Delta x_m
+#' }
+#'
+#' \deqn{
+#' \textrm{Stretch: }x = \alpha_m x_m
+#' }
+#'
+#' \deqn{
+#' \textrm{Shift + Stretch: } x = x_0 + \Delta x_m + \alpha_m (x_0 - <x_0>_m)
+#' }
+#'
+#' ## Shift
+#'
+#' If using a shift, for each month, a shift \eqn{\Delta x_m} is applied to
+#' \eqn{x_0}. \eqn{\Delta x_m} is the absolute change in the monthly mean value
+#' of the variable for the month \eqn{m},
+#' i.e. \eqn{\Delta x_m = <x_0>_m - <x>_m}. Here the monthly variance of the
+#' variable is unchanged.
+#'
+#' ## Stretch
+#'
+#' If using a stretch, for each month, a stretch \eqn{\alpha _m} is applied to
+#' \eqn{x_0}, where \eqn{\alpha _m} is the fractional change in the monthly-mean
+#' value of a variable, i.e. \eqn{\alpha _m = <x>_m / <x_0>_m}. In this case,
+#' the variance will be multiplied by to \eqn{\alpha^2_m}
+#'
+#' ## Combined Shift and Stretch
+#'
+#' When using a combined shift and stretch factor, both the mean and the
+#' variance will be switched off altogether.
+#'
+#' For more details about morphing, please see (Belcher etc. 2005)
+#'
 #' @param data An `epw_cmip6_data`object generated using [extract_data()]
 #'
 #' @param years An integer vector indicating the target years to be considered.
@@ -118,6 +195,10 @@ remove_units <- function (data, var) {
 #'        "shift")`. `"combined"` is only applicable to `tdb`. The default
 #'        morphing method for each variable is listed in the *Return* section.
 #'        If `NULL`, the default methods will be used. Default: `NULL`.
+#' @param warning If `TRUE`, warnings will be issued for cases with input data
+#'        less than a decade (10 years). This is because using data that only
+#'        covers a short period of time may not be able to capture the average
+#'        of future climate. Default: `FALSE`.
 #'
 #' @return An `epw_cmip6_morphed` object, which is basically a list of 12 elements:
 #'
@@ -136,22 +217,31 @@ remove_units <- function (data, var) {
 #' | 11   | `total_cover`  | [data.table::data.table()] | Derived         | Data of total sky cover after morphing                            |
 #' | 12   | `opaque_cover` | [data.table::data.table()] | Derived         | Data of opaque sky cover after morphing                           |
 #'
-#' Each [data.table::data.table()] listed above contains x columns
+#' Each [data.table::data.table()] listed above contains 19 columns below or an
+#' empty [data.table::data.table()] if the corresponding variables cannot be
+#' found in the input `epw_cmip6_data` object.
 #'
-#' | No.  | Column            | Type      | Description                                                                            |
-#' | ---: | -----             | -----     | -----                                                                                  |
-#' | 1    | `activity_drs`    | Character | Activity DRS (Data Reference Syntax)                                                   |
-#' | 2    | `institution_id`  | Character | Institution identifier                                                                 |
-#' | 3    | `source_id`       | Character | Model identifier                                                                       |
-#' | 4    | `experiment_id`   | Character | Root experiment identifier                                                             |
-#' | 5    | `member_id`       | Character | A compound construction from `sub_experiment_id` and `variant_label`                   |
-#' | 6    | `table_id`        | Character | Table identifier                                                                       |
-#' | 7    | `lat`             | Double    | The **averaged** values of input latitude                                              |
-#' | 8    | `lon`             | Double    | The **averaged** values of input longitude                                             |
-#' | 9    | `interval`        | Factor    | The year value of data morphed                                                         |
-#' | 10   | **Variable Name** | Double    | The morphed data, where `Variable Name` is the corresponding EPW weather variable name |
-#' | 11   | `delta`           | Double    | The shift factor. Will be `NA` for derivied values                                     |
-#' | 12   | `alpha`           | Double    | The stretch factor. Will be `NA` for derivied values                                   |
+#' | No.  | Column            | Type      | Description                                                                                           |
+#' | ---: | -----             | -----     | -----                                                                                                 |
+#' | 1    | `activity_drs`    | Character | Activity DRS (Data Reference Syntax)                                                                  |
+#' | 2    | `institution_id`  | Character | Institution identifier                                                                                |
+#' | 3    | `source_id`       | Character | Model identifier                                                                                      |
+#' | 4    | `experiment_id`   | Character | Root experiment identifier                                                                            |
+#' | 5    | `member_id`       | Character | A compound construction from `sub_experiment_id` and `variant_label`                                  |
+#' | 6    | `table_id`        | Character | Table identifier                                                                                      |
+#' | 7    | `lon`             | Double    | The **averaged** values of input longitude                                                            |
+#' | 8    | `lat`             | Double    | The **averaged** values of input latitude                                                             |
+#' | 9    | `dist`            | Double    | The **averaged** spherical distances in km between EPW location and grid coordinates                  |
+#' | 10   | `interval`        | Factor    | The label value used to average raw input data                                                        |
+#' | 11   | `datetime`        | POSIXct   | The datetime value with **fake year** generated by calling the `Epw$data()` method with the input EPW |
+#' | 12   | `year`            | Integer   | The **original** year of the raw EPW data                                                             |
+#' | 13   | `month`           | Integer   | The month value of the morphed data                                                                   |
+#' | 14   | `day`             | Integer   | The day of the morphed data                                                                           |
+#' | 15   | `hour`            | Integer   | The hour of the morphed data                                                                          |
+#' | 16   | `minute`          | Integer   | The minute of the morphed data                                                                        |
+#' | 17   | **Variable Name** | Double    | The morphed data, where `Variable Name` is the corresponding EPW weather variable name                |
+#' | 18   | `delta`           | Double    | The shift factor. Will be `NA` for derived values                                                    |
+#' | 19   | `alpha`           | Double    | The stretch factor. Will be `NA` for derived values                                                  |
 #'
 #' @references
 #' Belcher, S., Hacker, J., Powell, D., 2005. Constructing design weather data
@@ -159,7 +249,7 @@ remove_units <- function (data, var) {
 #' 26, 49–61. https://doi.org/10.1191/0143624405bt112oa
 #'
 #' @export
-morphing_epw <- function (data, years = NULL, labels = NULL, methods = NULL) {
+morphing_epw <- function (data, years = NULL, labels = NULL, methods = NULL, warning = FALSE) {
     assert_class(data, "epw_cmip6_data")
     if (is.null(methods)) {
         methods <- c(tdb = "stretch", rh = "stretch", p = "stretch",
@@ -189,7 +279,7 @@ morphing_epw <- function (data, years = NULL, labels = NULL, methods = NULL) {
         tasmin <- data_cmip[J("tasmin"), on = "variable", nomatch = NULL]
         if (!nrow(tasmax)) tasmax <- NULL
         if (!nrow(tasmin)) tasmin <- NULL
-        tdb <- morphing_tdb(data_epw, tas, tasmax, tasmin, years, labels = labels, type = methods["tdb"])
+        tdb <- morphing_tdb(data_epw, tas, tasmax, tasmin, years, labels = labels, type = methods["tdb"], warning = warning)
     }
 
     # NODE 8: RH
@@ -203,7 +293,7 @@ morphing_epw <- function (data, years = NULL, labels = NULL, methods = NULL) {
         hursmin <- data_cmip[J("hursmin"), on = "variable", nomatch = NULL]
         if (!nrow(hursmax)) hursmax <- NULL
         if (!nrow(hursmin)) hursmin <- NULL
-        rh <- morphing_rh(data_epw, hurs, hursmax, hursmin, years, labels = labels, type = methods["rh"])
+        rh <- morphing_rh(data_epw, hurs, hursmax, hursmin, years, labels = labels, type = methods["rh"], warning = warning)
     }
 
     # NODE 7: Tdew
@@ -222,7 +312,7 @@ morphing_epw <- function (data, years = NULL, labels = NULL, methods = NULL) {
         verbose("WARNING: Input does not contain any data of 'sea level pressure'. Skip.")
         p <- data.table()
     } else {
-        p <- morphing_pa(data_epw, psl, years, labels = labels, type = methods["p"])
+        p <- morphing_pa(data_epw, psl, years, labels = labels, type = methods["p"], warning = warning)
     }
 
     # NODE 10: Extraterrestrial direct normal radiation [NOT USED in EnergyPlus]
@@ -235,10 +325,10 @@ morphing_epw <- function (data, years = NULL, labels = NULL, methods = NULL) {
     )]
     rlds <- data_cmip[J("rlds"), on = "variable", nomatch = NULL]
     if (!nrow(rlds)) {
-        verbose("WARNING: Input does not contain any data of 'surface downwelling longware radiation'. Skip.")
+        verbose("WARNING: Input does not contain any data of 'surface downwelling longwave radiation'. Skip.")
         hor_ir <- data.table()
     } else {
-        hor_ir <- morphing_hor_ir(data_epw, rlds, years, labels = labels, type = methods["hor_ir"])
+        hor_ir <- morphing_hor_ir(data_epw, rlds, years, labels = labels, type = methods["hor_ir"], warning = warning)
     }
 
     # NODE 13: Global horizontal radiation
@@ -251,7 +341,7 @@ morphing_epw <- function (data, years = NULL, labels = NULL, methods = NULL) {
         verbose("WARNING: Input does not contain any data of 'surface downwelling shortware radiation'. Skip.")
         glob_rad <- data.table()
     } else {
-        glob_rad <- morphing_glob_rad(data_epw, rsds, years, labels = labels, type = methods["glob_rad"])
+        glob_rad <- morphing_glob_rad(data_epw, rsds, years, labels = labels, type = methods["glob_rad"], warning = warning)
     }
 
     #!NODE 15: Diffuse horizontal radiation
@@ -287,7 +377,7 @@ morphing_epw <- function (data, years = NULL, labels = NULL, methods = NULL) {
         verbose("WARNING: Input does not contain any data of 'near-surface wind speed'. Skip.")
         wind <- data.table()
     } else {
-        wind <- morphing_wind_speed(data_epw, sfcWind, years, labels = labels, type = methods["wind"])
+        wind <- morphing_wind_speed(data_epw, sfcWind, years, labels = labels, type = methods["wind"], warning = warning)
     }
 
     # NODE 22: Total sky cover
@@ -297,7 +387,7 @@ morphing_epw <- function (data, years = NULL, labels = NULL, methods = NULL) {
         verbose("WARNING: Input does not contain any data of 'total cloud area fraction for the whole atmospheric column'. Skip.")
         total_cover <- data.table()
     } else {
-        total_cover <- morphing_total_sky_cover(data_epw, clt, years, labels = labels)
+        total_cover <- morphing_total_sky_cover(data_epw, clt, years, labels = labels, warning = warning)
     }
 
     # NODE 23: Opaque sky cover
@@ -328,7 +418,7 @@ morphing_epw <- function (data, years = NULL, labels = NULL, methods = NULL) {
 # morphing_from_mean {{{
 morphing_from_mean <- function (var, data_epw, data_mean, data_max = NULL, data_min = NULL,
                                 type = c("shift", "stretch", "combined"),
-                                years = NULL, labels = NULL) {
+                                years = NULL, labels = NULL, warning = FALSE) {
     type <- match.arg(type)
 
     if (!nrow(data_mean)) return(data.table())
@@ -341,25 +431,25 @@ morphing_from_mean <- function (var, data_epw, data_mean, data_max = NULL, data_
 
     # add cut interval and average by lon, lat, month and day (i.e. monthly
     # avarage) in CMIP6 data
-    data_mean <- preprocess_morphing(data_mean, leapyear = FALSE, years = years, labels = labels)
+    data_mean <- preprocess_morphing(data_mean, leapyear = FALSE, years = years, labels = labels, warning = warning)
     # this will automatically do unit conversions like K --> C
     data_mean <- align_units(data_mean, u)
 
     if (type == "combined" && !is.null(data_max) && !is.null(data_min)) {
-        data_max <- preprocess_morphing(data_max, leapyear = FALSE, years = years, labels = labels)
-        data_min <- preprocess_morphing(data_min, leapyear = FALSE, years = years, labels = labels)
+        data_max <- preprocess_morphing(data_max, leapyear = FALSE, years = years, labels = labels, warning = warning)
+        data_min <- preprocess_morphing(data_min, leapyear = FALSE, years = years, labels = labels, warning = warning)
 
         # this will automatically do unit conversions like K --> C
         data_max <- align_units(data_max, u)
         data_min <- align_units(data_min, u)
 
         # merge max and min CMIP6 values into mean data.table
-        data_mean[data_max, on = c("activity_drs", "experiment_id", "institution_id",
-            "source_id", "member_id", "table_id", "lat", "lon", "units", "month", "interval"),
+        data_mean[data_max, on = c("activity_drs", "institution_id", "source_id",
+            "experiment_id", "member_id", "table_id", "lat", "lon", "dist", "units", "month", "interval"),
             value_max := i.value
         ]
-        data_mean[data_min, on = c("activity_drs", "experiment_id", "institution_id",
-            "source_id", "member_id", "table_id", "lat", "lon", "units", "month", "interval"),
+        data_mean[data_min, on = c("activity_drs", "institution_id", "source_id",
+            "experiment_id", "member_id", "table_id", "lat", "lon", "dist", "units", "month", "interval"),
             value_min := i.value
         ]
 
@@ -367,24 +457,37 @@ morphing_from_mean <- function (var, data_epw, data_mean, data_max = NULL, data_
         # Otherwise the final results will be NA
         # NOTE: By doing so, alpha for those GCMS will be zero and 'shift'
         # method is used. Warnings should be issued
-        i_max <- data_mean[J(NA_real_), on = "value_max", which = TRUE]
-        i_min <- data_mean[J(NA_real_), on = "value_min", which = TRUE]
+        i_max <- data_mean[J(NA_real_), on = "value_max", which = TRUE, nomatch = NULL]
+        i_min <- data_mean[J(NA_real_), on = "value_min", which = TRUE, nomatch = NULL]
         i <- unique(c(i_min, i_max))
 
-        # construct case string
-        cases <- data_mean[i, unique(sprintf("CMIP6.%s.%s.%s.%s.%s.%s",
-            activity_drs, institution_id, source_id, experiment_id, member_id, table_id))]
-        cases <- sprintf("[%i] '%s'", seq_along(cases), sort(cases))
-        # issue warnings
-        warning(sprintf("Case(s) below does not contains max or min of '%s' data. ", gsub("_", " ", var)),
-            "'Shift' method will be used for it.\n", paste0(cases, collapse = "\n"),
-            call. = FALSE
-        )
+        case_fallback <- data.table()
+        if (length(i)) {
+            cols <- c("activity_drs", "institution_id", "source_id",
+                "experiment_id", "member_id", "table_id")
 
-        set(data_mean, i_max, "value_max", data_mean$value[i_max])
-        set(data_mean, i_min, "value_min", data_mean$value[i_max])
+            case_fallback <- unique(data_mean[i], by = cols)
+            set(case_fallback, NULL, setdiff(names(case_fallback), cols), NULL)
+
+            # construct case string
+            cases <- case_fallback[, unique(sprintf("CMIP6.%s.%s.%s.%s.%s.%s",
+                activity_drs, institution_id, source_id, experiment_id, member_id, table_id))]
+            cases <- sprintf("[%i] '%s'", seq_along(cases), sort(cases))
+            # issue warnings
+            warning(sprintf("Case(s) below contains missing values of max or min of '%s' data. ", gsub("_", " ", var)),
+                "'Shift' method will be used for it.\n", paste0(cases, collapse = "\n"),
+                call. = FALSE
+            )
+        }
     }
 
+    # calculate delta, alpha and add EPW monthly average value
+    data_mean[monthly, on = "month", `:=`(
+        delta = value - val_mean, alpha = value / val_mean,
+        epw_mean = i.val_mean,
+        epw_max = i.val_max, epw_min = i.val_min
+    )]
+    
     # add datetime columns from the original EPW data into the monthly average of
     # CMIP6 data
     # after this every row in 'data' indicates a specific hour (as EPW has
@@ -392,17 +495,50 @@ morphing_from_mean <- function (var, data_epw, data_mean, data_max = NULL, data_
     data <- data_epw[, .SD, .SDcols = c("datetime", "year", "month", "day", "hour", "minute", var)][
         data_mean, on = "month", allow.cartesian = TRUE]
 
-    # calculate delta, alpha and add EPW monthly average value
-    data[monthly, on = "month", `:=`(delta = value - i.val_mean,
-        epw_mean = i.val_mean, epw_max = i.val_max, epw_min = i.val_min
-    )]
-
     if (type == "combined" && all(c("value_min", "value_max") %in% names(data))) {
         data[, alpha := ((value_max - epw_max) - (value_min - epw_min)) / (epw_max - epw_min)]
+        if (nrow(case_fallback)) {
+            data[case_fallback, on = c(names(case_fallback)), alpha := 0.0]
+        }
     } else {
         data[, alpha := value / epw_mean]
     }
 
+    # check the alpha values
+    # if values are too small or too large, issue warnings and fall back to
+    # shift method
+    thres_alpha <- getOption("epwshiftr.threshold_alpha")
+    if (!checkmate::test_number(thres_alpha, lower = 0)) {
+        warning(paste0(
+            "The threshold value for the monthly-mean fractional change (Alpha) ",
+            "should be a positive number, but '",
+            if (is.null(thres_alpha)) "NULL" else thres_alpha,
+            "' is found."
+        ))
+    }
+    if (type %in% c("stretch", "combined") && nrow(abnorm_alpha <- data_mean[abs(units::drop_units(alpha)) > thres_alpha])) {
+        warning(sprintf(
+            paste(
+                "The absolute values of monthly-mean fractional change (Alpha) below",
+                "for '%s' has exceeded the threshold (%s) set by the option",
+                "'epwshiftr.threshold_alpha'. 'Shift' morphing method will be utilized",
+                "instead of '%s' method to avoid unrealistic values. It is highly",
+                "suggested to further investigate the input data.\n%s",
+                collapse = " "
+            ),
+            gsub("_", " ", var, fixed = TRUE), thres_alpha, type,
+            paste0(sprintf(
+                "Month = %s | Monthly-mean: EPW = %s, GCM = %s --> Alpha = %s",
+                format(abnorm_alpha$month),
+                format(abnorm_alpha$epw_mean, digits = 3),
+                format(abnorm_alpha$value, digits = 3),
+                format(units::drop_units(abnorm_alpha$alpha), digits = 3)
+            ), collapse = "\n")
+        ))
+        
+        type <- "shift"
+    }
+
     if (type == "shift") {
         data[, c(var) := units::set_units(get(var) + delta, u, mode = "standard")]
     } else if (type == "stretch") {
@@ -417,8 +553,8 @@ morphing_from_mean <- function (var, data_epw, data_mean, data_max = NULL, data_
 
     data[, .SD, .SDcols = c(
         # meta from CMIP6
-        "activity_drs", "experiment_id", "institution_id", "source_id", "member_id",
-        "table_id", "lon", "lat",
+        "activity_drs", "institution_id", "source_id", "experiment_id", "member_id",
+        "table_id", "lon", "lat", "dist",
         # interval
         "interval",
         # datetime
@@ -431,7 +567,7 @@ morphing_from_mean <- function (var, data_epw, data_mean, data_max = NULL, data_
 # }}}
 
 # morphing_tdb {{{
-morphing_tdb <- function (data_epw, tas, tasmax = NULL, tasmin = NULL, years = NULL, labels = NULL, type = "combined") {
+morphing_tdb <- function (data_epw, tas, tasmax = NULL, tasmin = NULL, years = NULL, labels = NULL, type = "combined", warning = FALSE) {
     morphing_from_mean(
         var = "dry_bulb_temperature",
         data_epw = data_epw,
@@ -440,13 +576,14 @@ morphing_tdb <- function (data_epw, tas, tasmax = NULL, tasmin = NULL, years = N
         data_min = tasmin,
         years = years,
         labels = labels,
-        type = type
+        type = type,
+        warning = warning
     )
 }
 # }}}
 
 # morphing_rh {{{
-morphing_rh <- function (data_epw, hurs, hursmax = NULL, hursmin = NULL, years = NULL, labels = NULL, type = "combined") {
+morphing_rh <- function (data_epw, hurs, hursmax = NULL, hursmin = NULL, years = NULL, labels = NULL, type = "combined", warning = FALSE) {
     rh <- morphing_from_mean(
         var = "relative_humidity",
         data_epw = data_epw,
@@ -455,7 +592,8 @@ morphing_rh <- function (data_epw, hurs, hursmax = NULL, hursmin = NULL, years =
         data_min = hursmin,
         years = years,
         labels = labels,
-        type = type
+        type = type,
+        warning = warning
     )
 
     # reset RH > 100% to 100%
@@ -499,40 +637,43 @@ morphing_tdew <- function (tdb, rh) {
 # }}}
 
 # morphing_pa {{{
-morphing_pa <- function (data_epw, psl, years = NULL, labels = NULL, type = "stretch") {
+morphing_pa <- function (data_epw, psl, years = NULL, labels = NULL, type = "stretch", warning = FALSE) {
     morphing_from_mean(
         var = "atmospheric_pressure",
         data_epw = data_epw,
         data_mean = psl,
         years = years,
         labels = labels,
-        type = type
+        type = type,
+        warning = warning
     )
 }
 # }}}
 
 # morphing_hor_ir {{{
-morphing_hor_ir <- function (data_epw, rlds, years = NULL, labels = NULL, type = "stretch") {
+morphing_hor_ir <- function (data_epw, rlds, years = NULL, labels = NULL, type = "stretch", warning = FALSE) {
     morphing_from_mean(
         var = "horizontal_infrared_radiation_intensity_from_sky",
         data_epw = data_epw,
         data_mean = rlds,
         years = years,
         labels = labels,
-        type = type
+        type = type,
+        warning = warning
     )
 }
 # }}}
 
 # morphing_glob_rad {{{
-morphing_glob_rad <- function (data_epw, rsds, years = NULL, labels = NULL, type = "stretch") {
+morphing_glob_rad <- function (data_epw, rsds, years = NULL, labels = NULL, type = "stretch", warning = FALSE) {
     morphing_from_mean(
         var = "global_horizontal_radiation",
         data_epw = data_epw,
         data_mean = rsds,
         years = years,
         labels = labels,
-        type = type
+        type = type,
+        warning = warning
     )
 }
 # }}}
@@ -565,23 +706,24 @@ morphing_norm_rad <- function (glob_rad, diff_rad) {
 # }}}
 
 # morphing_wind_speed {{{
-morphing_wind_speed <- function (data_epw, sfcWind, years = NULL, labels = NULL, type = "stretch") {
+morphing_wind_speed <- function (data_epw, sfcWind, years = NULL, labels = NULL, type = "stretch", warning = FALSE) {
     morphing_from_mean(
         var = "wind_speed",
         data_epw = data_epw,
         data_mean = sfcWind,
         years = years,
         labels = labels,
-        type = type
+        type = type,
+        warning = warning
     )
 }
 # }}}
 
 # morphing_total_sky_cover {{{
-morphing_total_sky_cover <- function (data_epw, clt, years = NULL, labels = NULL) {
+morphing_total_sky_cover <- function (data_epw, clt, years = NULL, labels = NULL, warning = FALSE) {
     var <- "total_sky_cover"
     if (!nrow(clt)) return(data.table())
-    data_mean <- preprocess_morphing(clt, leapyear = FALSE, years = years, labels = labels)
+    data_mean <- preprocess_morphing(clt, leapyear = FALSE, years = years, labels = labels, warning = warning)
     monthly <- unique(data_epw[, .SD, .SDcols = c("month")])
 
     data_mean <- data_mean[monthly, on = c("month")]
@@ -599,7 +741,7 @@ morphing_total_sky_cover <- function (data_epw, clt, years = NULL, labels = NULL
     data[, .SD, .SDcols = c(
         # meta from CMIP6
         "activity_drs", "experiment_id", "institution_id", "source_id", "member_id",
-        "table_id", "lon", "lat",
+        "table_id", "lon", "lat", "dist",
         # interval
         "interval",
         # datetime
@@ -621,8 +763,8 @@ morphing_opaque_sky_cover <- function (data_epw, total_sky_cover) {
 
     data[, .SD, .SDcols = c(
         # meta from CMIP6
-        "activity_drs", "experiment_id", "institution_id", "source_id", "member_id",
-        "table_id", "lon", "lat",
+        "activity_drs", "institution_id", "source_id", "experiment_id", "member_id",
+        "table_id", "lon", "lat", "dist",
         # interval
         "interval",
         # datetime
@@ -633,23 +775,26 @@ morphing_opaque_sky_cover <- function (data_epw, total_sky_cover) {
 }
 # }}}
 
+# nocov start
 # morphing_precipitation {{{
-morphing_precipitation <- function (data_epw, pr, years = NULL, labels = NULL, type = "stretch") {
+morphing_precipitation <- function (data_epw, pr, years = NULL, labels = NULL, type = "stretch", warning = FALSE) {
     morphing_from_mean(
         var = "precipitable_water",
         data_epw = data_epw,
         data_mean = pr,
         years = years,
         labels = labels,
-        type = type
+        type = type,
+        warning = warning
     )
 }
 # }}}
+# nocov end
 
 # location_mean {{{
 location_mean <- function (dt, by_exclude = NULL) {
-    res <- dt[, list(lon = mean(lon), lat = mean(lat), value = mean(value)),
-        by = c(setdiff(names(dt), c("lon", "lat", "value", by_exclude)))]
+    res <- dt[, list(lon = mean(lon), lat = mean(lat), dist = mean(dist), value = mean(value)),
+        by = c(setdiff(names(dt), c("lon", "lat", "dist", "value", by_exclude)))]
     data.table::setcolorder(res, setdiff(names(dt), by_exclude))[]
 }
 # }}}
@@ -660,12 +805,6 @@ to_radian <- function (degree) {
 }
 # }}}
 
-# to_degree {{{
-to_degree <- function (radian) {
-    radian * 180 / pi
-}
-# }}}
-
 # day_angle {{{
 day_angle <- function (day_of_year) {
     day_of_year * 360.0 / 365.25
@@ -708,38 +847,12 @@ solar_angle <- function (latitude, longitude, day_of_year, hour, timezone) {
 }
 # }}}
 
-# append_epw_data {{{
-# Append EPW data in each case for merely comparison purpose
-append_epw_data <- function (morphed) {
-    epw <- morphed$epw
-    epw$drop_unit()
-    epw <- epw$data()
-
-    mor <- morphed[names(morphed) != "epw"]
-
-    lapply(mor, function (dt) {
-        # get variable name
-        var <- intersect(names(dt), names(epw)[-(1:7)])
-
-        meta <- unique(dt[, .SD, .SDcols = c("experiment_id", "institution_id",
-            "source_id", "member_id", "table_id", "lat", "lon", "interval")])
-
-        base <- meta[, as.list(epw[, .SD, .SDcols = c(names(epw)[1:6], var)]), by = c(names(meta))]
-
-        set(base, NULL, "interval", "EPW")
-
-        # combine
-        rbindlist(list(dt, base), fill = TRUE)
-    })
-}
-# }}}
-
 # future_epw {{{
 #' Create future EPW files using morphed data
 #'
 #' @param morphed An `epw_cmip6_morphed` object created using [morphing_epw()].
 #' @param by A character vector of columns to be used as grouping variables when
-#'        creating EPW files. Should be a subeset of:
+#'        creating EPW files. Should be a subset of:
 #'
 #' * `"experiment"`: root experiment identifiers
 #' * `"source"`: model identifiers
@@ -758,14 +871,26 @@ append_epw_data <- function (morphed) {
 #'        using grouping variables specified in `by`.
 #' @param overwrite If `TRUE`, overwrite existing files if they exist. Default:
 #'        `FALSE`.
+#' @param full If `TRUE`, a [data.table][data.table::data.table()] containing
+#'        information about how the data are split and also the generated future
+#'        EPWs and their paths are returned. Default: `FALSE`.
+#'
+#' @return
+#'
+#' If `full` is `FALSE`, which is the default, a list of generated [eplusr::Epw] objects, invisibly.
+#' Otherwise, a [data.table][data.table::data.table()] with columns:
+#'
+#' * specified by the `by` value
+#' * `epw`: a list of [eplusr::Epw]
+#' * `path`: full paths of the generated EPW files
 #'
-#' @return A list of generated [eplusr::Epw] objects, invisibly
 #' @export
 future_epw <- function (morphed, by = c("experiment", "source", "interval"),
-                        dir = ".", separate = TRUE, overwrite = FALSE) {
+                        dir = ".", separate = TRUE, overwrite = FALSE, full = FALSE) {
     assert_class(morphed, "epw_cmip6_morphed")
     assert_string(dir)
     assert_flag(separate)
+    assert_flag(full)
 
     epw <- morphed$epw
 
@@ -800,7 +925,7 @@ future_epw <- function (morphed, by = c("experiment", "source", "interval"),
     }
 
     if (!length(morphed)) {
-        stop("No morphed data found. Please run 'morphing_epw' first.")
+        stop("No morphed data found. Please run 'morphing_epw()' first.")
     }
 
     # remove delta and alpha columns
@@ -858,9 +983,11 @@ future_epw <- function (morphed, by = c("experiment", "source", "interval"),
     dir <- normalizePath(dir, mustWork = FALSE)
     if (!dir.exists(dir)) dir.create(dir, showWarnings = FALSE, recursive = TRUE)
     if (!checkmate::test_directory(dir)) {
+        # nocov start
         stop(sprintf("Failed to create output directory '%s'"),
             normalizePath(dir, mustWork = FALSE)
         )
+        # nocov end
     }
 
     if (separate) {
@@ -870,7 +997,7 @@ future_epw <- function (morphed, by = c("experiment", "source", "interval"),
         output <- file.path(dir, fn)
     }
 
-    invisible(lapply(seq_along(output), function (i) {
+    epws <- lapply(seq_along(output), function (i) {
         # clone the original EPW
         new_epw <- epw$clone()
 
@@ -887,14 +1014,28 @@ future_epw <- function (morphed, by = c("experiment", "source", "interval"),
         new_dir <- dirname(output[i])
         if (!dir.exists(new_dir)) dir.create(new_dir, showWarnings = FALSE, recursive = TRUE)
         if (!checkmate::test_directory(new_dir)) {
+            # nocov start
             stop(sprintf("Failed to create output directory '%s'"),
                 normalizePath(new_dir, mustWork = FALSE)
             )
+            # nocov end
         }
         new_epw$save(output[i], overwrite = overwrite)
 
         new_epw
-    }))
+    })
+
+    if (!full) return(invisible(epws))
+
+    meta <- rbindlist(lapply(spl, function(d) d[1L, .SD, .SDcols = cols_by]))
+
+    # rename the columns
+    setnames(meta, cols_by, names(dict)[match(cols_by, names(dict))])
+
+    set(meta, NULL, "epw", epws)
+    set(meta, NULL, "path", vapply(epws, function(epw) normalizePath(epw$path()), character(1)))
+
+    meta
 }
 # }}}
 
