@@ -6,7 +6,7 @@ get_nc_meta <- function (file) {
     # get all attributes
     atts <- get_nc_atts(file)
 
-    # get meta data
+    # get metadata
     meta <- as.list(atts[
         J("NC_GLOBAL",
           c("mip_era", "activity_id", "institution_id", "source_id",
@@ -39,21 +39,36 @@ get_nc_meta <- function (file) {
 # summary_database {{{
 #' Summary CMIP6 model output file status
 #'
-#' `summary_database()` scan the directory specified and returns a
+#' `summary_database()` scans the directory specified and returns a
 #' [data.table()] containing summary information about all the CMIP6
 #' files available against the output file index loaded using
 #' [load_cmip6_index()].
 #'
-#' `summary_database()` uses [future.apply][future.apply::future_lapply()]
-#' underneath. You can use your preferable future backend to
-#' speed up data extraction in parallel. By default, `summary_database()` uses
-#' `future::sequential` backend, which runs things in sequential.
+#' The database here can be any directory that stores the NetCDF files for CMIP6
+#' GCMs. It can be also be the same as `get_data_dir()` where epwshiftr stores
+#' the output file index, if you want to save the output file index and output
+#' files in the same place.
 #'
-#' @param dir A single string indcating the directory where CMIP6 model output
+#' `summary_database()` uses the `tracking_id`, `datetime_start` and
+#' `datetime_end` global attributes of each NetCDF file to match against the
+#' output file index. So the names of NetCDF files do not necessarily follow the
+#' CMIP6 file name encoding.
+#'
+#' `summary_database()` will append 5 columns in the CMIP6 output file index:
+#'
+#' - `file_path`: the full path of matched NetCDF file for every case.
+#'
+#' `summary_database()` uses [future.apply][future.apply::future_lapply()]
+#' underneath to speed up the data processing if applicable. You can use your
+#' preferable future backend to speed up data extraction in parallel. By default,
+#' `summary_database()` uses `future::sequential` backend, which runs things in
+#' sequential.
+#'
+#' @param dir A single string indicating the directory where CMIP6 model output
 #'        NetCDF files are stored.
 #'
 #' @param by The grouping column to summary the database status. Should be a
-#'        subeset of:
+#'        subset of:
 #'
 #' * `"experiment"`: root experiment identifiers
 #' * `"source"`: model identifiers
@@ -72,6 +87,13 @@ get_nc_meta <- function (file) {
 #'        will be used. If `"skip"`, all matched files will be skip and this
 #'        case will be kept as unmatched. Default: `"skip"`.
 #'
+#' @param miss Actions when matched files in the previous summary do not exist
+#'        when running current summary. Only applicable when `append` is set to
+#'        `TRUE`. If `"keep"`, the metadata for the missing output files will
+#'        be kept. If `"overwrite"`, existing metadata of those output will be
+#'        first removed from the output file index and overwritten based on the
+#'        newly matched files if possible. Default: `"keep"`.
+#'
 #' @param recursive If `TRUE`, scan recursively into directories. Default:
 #'        `FALSE`.
 #'
@@ -81,8 +103,8 @@ get_nc_meta <- function (file) {
 #'        database file saved in [get_data_dir()] will remain unchanged.
 #'        Default: `FALSE`.
 #'
-#' @param warning If `TRUE`, warning messages will show when multiple files
-#'        match a same case. Default: `TRUE`.
+#' @param warning If `TRUE`, warning messages will show when target files are
+#'        missing or multiple files match a same case. Default: `TRUE`.
 #'
 #' @return A [data.table::data.table()] containing corresponding grouping
 #' columns plus:
@@ -97,9 +119,16 @@ get_nc_meta <- function (file) {
 #' | `dl_percent`     | Units (%)      | Total percentage of file downloaded |
 #' | `dl_size`        | Units (Mbytes) | Total size of file downloaded       |
 #'
-#' Also an attribute `not_matched` is added to the returned
-#' [data.table::data.table()] which contains meta data for those CMIP6 output
-#' files that are not covered by current CMIP6 output file index.
+#' Also 2 extra [data.table::data.table()] are attached as **attributes**:
+#'
+#' - `not_found`: A [data.table::data.table()] that contains metadata for those
+#'   CMIP6 outputs that are listed in current CMIP6 output file index but the
+#'   existing file paths are not valid now and cannot be found in current
+#'   database.
+#'
+#' - `not_matched`: A [data.table::data.table()] that contains metadata for
+#'   those CMIP6 output files that are found in current database but not listed
+#'   in current CMIP6 output file index.
 #'
 #' For the meaning of grouping columns, see [init_cmip6_index()].
 #'
@@ -109,13 +138,15 @@ get_nc_meta <- function (file) {
 #'
 #' summary_database(by = "experiment")
 #' }
+#'
 #' @importFrom future.apply future_lapply
 #' @importFrom progressr with_progress
 #' @export
+# TODO: should all not_matched and not_found merged int one?
 summary_database <- function (
     dir,
     by = c("activity", "experiment", "variant", "frequency", "variable", "source", "resolution"),
-    mult = c("skip", "latest"), append = FALSE,
+    mult = c("skip", "latest"), append = FALSE, miss = c("keep", "overwrite"),
     recursive = FALSE, update = FALSE, warning = TRUE)
 {
     # column names
@@ -127,14 +158,15 @@ summary_database <- function (
     assert_subset(by, choices = dict)
 
     mult <- match.arg(mult)
+    miss <- match.arg(miss)
 
     # load index
     idx <- load_cmip6_index()
 
     # find all nc files in specified directory
-    ncfiles <- list.files(dir, "\\.nc$", full.names = TRUE, recursive = recursive)
+    ncfiles <- list.files(dir, "[.](nc)|(hdf)$", full.names = TRUE, recursive = recursive)
 
-    message(paste0("", length(ncfiles), " NetCDF files found."))
+    verbose(paste0("", length(ncfiles), " NetCDF files found."))
 
     # columns to be added
     cols <- c("file_path", "file_realsize", "file_mtime", "time_units", "time_calendar")
@@ -152,6 +184,9 @@ summary_database <- function (
                 set(idx, NULL, nm, vals[[nm]])
             }
         }
+
+        left <- data.table()
+        miss <- data.table()
     } else {
         progressr::with_progress({
             p <- progressr::progressor(along = ncfiles)
@@ -179,13 +214,63 @@ summary_database <- function (
         # NOTE: Should use right join here instead of adding by reference. This
         # is because for some GCMs, tracking id can be the same for multiple
         # files. The most safe way is to add additional checking for datetime
-        # a) first match using tracking id
+        #
+        # a) remove existing output file metadata if necessary
+        if (!append && any(cols %in% cols_idx)) {
+            set(idx, NULL, cols[cols %in% cols_idx], NULL)
+        }
+
+        # b) check existence of previous matched files
+        is_lost <- FALSE # in case append is set to FALSE
+        if (append && "file_path" %in% cols_idx && any(is_lost <- !is.na(idx$file_path) & !file.exists(idx$file_path))) {
+            lost <- idx[is_lost]
+            if (miss == "overwrite") {
+                i_lost <- which(is_lost)
+                set(idx, i_lost, "file_path", NA_character_)
+                set(idx, i_lost, "file_realsize", NA_real_)
+                set(idx, i_lost, "file_mtime", as.POSIXct(NA))
+                set(idx, i_lost, "time_units", NA_character_)
+                set(idx, i_lost, "time_calendar", NA_character_)
+            }
+
+            if (warning) {
+                # construct error message:
+                lost[, index_case := data.table::rleid(index)]
+                mes <- lost[, by = "index_case", {
+                    head <- sprintf("#%i | For case '%s':\n", .BY$index_case, gsub("\\|.+$", "", file_id[1]))
+                    file <- sprintf("   --> file '%s'", normalizePath(file_path, mustWork = FALSE))
+                    list(message = paste0(head, file))
+                }]$message
+                set(lost, NULL, "index_case", NULL)
+
+                ori <- getOption("warning.length")
+                options(warning.length = 8170L)
+                on.exit(options(warning.length = ori), add = TRUE)
+
+                warning("Previously matched NetCDF file(s) below does not exist anymore. ",
+                    if (miss == "keep") {
+                        "Its metadata are kept in the output file index as `miss` is set to `\"keep\"`.\n"
+                    } else if (miss == "overwrite") {
+                        paste(
+                            "Since `miss` is set to `\"overwrite\"`, its metadata has been **removed** from the output file index",
+                            "and will be overwritten by the data from new matched NetCDF file if possible.\n"
+                        )
+                    },
+                    paste0(mes, collapse = "\n"),
+                    "\n\nYou can run `attr(x, \"not_found\")` to see the metadata of those files.",
+                    call. = FALSE
+                )
+            }
+            set(lost, NULL, "index", NULL)
+        }
+
+        # c) first match using tracking id
         idx_m <- ncmeta[, .SD, .SDcols = c("tracking_id", "datetime_start", "datetime_end", cols)][idx, on = "tracking_id", nomatch = NULL]
 
-        # b) remove files that do not have any overlap in terms of datetime range
+        # d) remove files that do not have any overlap in terms of datetime range
         idx_m <- idx_m[!(datetime_start > i.datetime_end | datetime_end < i.datetime_start)]
 
-        # c) calculate the overlapped percentages coverred datetime of input
+        # e) calculate the overlapped percentages coverred datetime of input
         # file to index data
         set(idx_m, NULL, "index_match", seq_len(nrow(idx_m)))
         if (!nrow(idx_m)) {
@@ -199,12 +284,17 @@ summary_database <- function (
             }]
         }
 
-        # d) only keep items that have overlapped percentages larger than 60%
+        # f) only keep items that have overlapped percentages larger than 60%
         idx_m <- idx_m[overlap >= 0.6]
         set(idx_m, NULL, c("overlap", "index_match"), NULL)
 
-        # e) keep rows whose files have not yet been found
-        idx <- rbindlist(list(idx[!idx_m, on = "index"], idx_m), fill = TRUE)
+        # g) keep rows whose files have not yet been found
+        if (!any(is_lost)) {
+            idx <- rbindlist(list(idx[!idx_m, on = "index"], idx_m), fill = TRUE)
+        } else {
+            should_keep <- is_lost | !idx$index %in% idx_m$index
+            idx <- rbindlist(list(idx[should_keep], idx_m[J(idx$index[!should_keep]), on = "index"]), fill = TRUE)
+        }
         data.table::setorderv(idx, "index")
 
         # store files that are not matched
@@ -216,16 +306,18 @@ summary_database <- function (
             # remove original file mata columns
             set(idx, NULL, orig, NULL)
         } else {
-            # keep original match
-            idx[J(NA_character_), on = "i.file_path", `:=`(
-                i.file_path = file_path,
-                i.file_realsize = file_realsize,
-                i.file_mtime = file_mtime,
-                i.time_units = time_units,
-                i.time_calendar = time_calendar,
-                i.datetime_start = datetime_start,
-                i.datetime_end = datetime_end
-            )]
+            if ("i.file_path" %in% orig) {
+                # keep original match
+                idx[J(NA_character_), on = "i.file_path", `:=`(
+                    i.file_path = file_path,
+                    i.file_realsize = file_realsize,
+                    i.file_mtime = file_mtime,
+                    i.time_units = time_units,
+                    i.time_calendar = time_calendar,
+                    i.datetime_start = datetime_start,
+                    i.datetime_end = datetime_end
+                )]
+            }
             new <- gsub("^i\\.", "", orig)
             set(idx, NULL, new, NULL)
             setnames(idx, orig, new)
@@ -269,19 +361,52 @@ summary_database <- function (
                     "Please check if there are duplicated NetCDF files in your database ",
                     "and check NetCDF global attributes of those files ",
                     "to see if there are any incorrect values. ", act, "\n",
-                    paste0(mes, collapse = "\n"), call. = FALSE)
+                    paste0(mes, collapse = "\n"),
+                    "\nYou can run `attr(x, \"not_matched\")` to see the metadata of those files.",
+                    call. = FALSE)
             }
         }
 
+        # keep the original order
+        setcolorder(idx, setdiff(cols_idx, c("index", cols)))
+
+        # store files that are not found after updating
+        miss <- idx[is.na(file_path)]
+        set(miss, NULL, cols, NULL)
+
+        if (nrow(miss) && warning) {
+            # construct error message:
+            miss[, index_case := data.table::rleid(index)]
+            mes <- miss[, by = "index_case", {
+                list(message = sprintf("#%i | For case '%s':\n", .BY$index_case, gsub("\\|.+$", "", file_id[1])))
+            }]$message
+            set(miss, NULL, "index_case", NULL)
+
+            ori <- getOption("warning.length")
+            options(warning.length = 8170L)
+            on.exit(options(warning.length = ori), add = TRUE)
+
+            warning("Case(s) shown below does not matche any NetCDF file in the database. ",
+                "Please make sure all needed NetCDF files listed in the file index have been downloaded and placed in the database.\n",
+                paste0(mes, collapse = "\n"),
+                "\n\nYou can run `attr(x, \"not_found\")` to see the metadata of those files.",
+                call. = FALSE)
+        }
+
         # remove index
+        set(miss, NULL, "index", NULL)
         set(idx, NULL, "index", NULL)
 
-        # keep the original order
-        setcolorder(idx, setdiff(cols_idx, "index"))
+        # combine not found and not matched together
+        if (any(is_lost)) {
+            # in this case, file_path, file_realsize and other columns are added
+            # for missed data
+            miss <- rbindlist(list(miss, lost), fill = TRUE)
+        }
     }
 
     # update index
-    EPWSHIFTR_ENV$index_db <- copy(idx)
+    this$index_db <- copy(idx)
 
     if (update) {
         # save database into the app data directory
@@ -302,6 +427,7 @@ summary_database <- function (
     ), by = c(by_cols)][order(-dl_percent)]
 
     setattr(sm, "not_matched", left)
+    setattr(sm, "not_found", miss)
 
     sm
 }
@@ -311,7 +437,7 @@ summary_database <- function (
 #' @importFrom data.table as.data.table set setattr setcolorder
 #' @importFrom RNetCDF utcal.nc
 #' @importFrom units set_units
-get_nc_data <- function (x, lats, lons, years, unit = TRUE) {
+get_nc_data <- function (x, coord, years, unit = TRUE) {
     assert_flag(unit)
     assert_integerish(years, lower = 1900, unique = TRUE, sorted = TRUE, any.missing = FALSE, null.ok = TRUE)
 
@@ -336,15 +462,12 @@ get_nc_data <- function (x, lats, lons, years, unit = TRUE) {
 
     # match time
     time <- match_nc_time(nc, years)
-
-    which_lat <- lats$which
-    which_lon <- lons$which
     which_time <- time$which
 
     # get dimensions
     dims <- get_nc_dims(nc)[J(c("lon", "lat", "time")), on = "name"][order(id)][, id := .I]
-    dims[J("lon"), on = "name", `:=`(value = list(which_lon))]
-    dims[J("lat"), on = "name", `:=`(value = list(which_lat))]
+    dims[J("lon"), on = "name", `:=`(value = list(coord$ind_lon))]
+    dims[J("lat"), on = "name", `:=`(value = list(coord$ind_lat))]
 
     # find correct dimentions
     ord <- list(
@@ -362,7 +485,9 @@ get_nc_data <- function (x, lats, lons, years, unit = TRUE) {
                 if (grepl("exceeds dimension bound", conditionMessage(e), fixed = TRUE)) {
                     NULL
                 } else {
+                    # nocov start
                     signalCondition(e)
+                    # nocov end
                 }
             }
         ))
@@ -372,9 +497,9 @@ get_nc_data <- function (x, lats, lons, years, unit = TRUE) {
     dt <- rbindlist(use.names = TRUE, lapply(seq_along(which_time), function (i) {
         if (!length(which_time[[i]])) {
             return(data.table(
-                value = double(),
-                lon = double(), lat = double(),
-                datetime = Sys.time()[0]
+                index = integer(),
+                lon = double(), lat = double(), dist = double(),
+                datetime = Sys.time()[0], value = double()
             ))
         }
 
@@ -386,16 +511,27 @@ get_nc_data <- function (x, lats, lons, years, unit = TRUE) {
         # longitude, latitude, time).
         dt <- as.data.table(RNetCDF::var.get.nc(nc, var,
             c(min(dims$value[[1L]]), min(dims$value[[2L]]), min(dims$value[[3L]])),
-            c(length(dims$value[[1L]]), length(dims$value[[2L]]), length(dims$value[[3L]])),
+            c(length(unique(dims$value[[1L]])), length(unique(dims$value[[2L]])), length(dims$value[[3L]])),
             collapse = FALSE)
         )
 
-        setnames(dt, c(dims$name, "value"))
-        setnames(dt, "time", "datetime")
-
-        set(dt, NULL, "lon", lons$value[dt$lon])
-        set(dt, NULL, "lat", lats$value[dt$lat])
+        setnames(dt, c(paste0("ord_", dims$name), "value"))
+        setnames(dt, "ord_time", "datetime")
         set(dt, NULL, "datetime", time$datetime[[i]][dt$datetime])
+
+        # only keep the matched points
+        ind_lon <- ind_lat <- ord_lon <- ord_lat <- .GRP <- NULL # to make CRAN check happy
+        # NOTE: make a copy of input coord here, otherwise there will be a shallow
+        # copy warning
+        # should be cheap to do so, since the coord is normally a small DT
+        co <- copy(coord)
+        co[order(ind_lon), ord_lon := .GRP, by = "ind_lon"]
+        co[order(ind_lat), ord_lat := .GRP, by = "ind_lat"]
+        res <- dt[co, on = c("ord_lon", "ord_lat")]
+        set(res, NULL, c("ord_lon", "ord_lat", "ind_lon", "ind_lat"), NULL)
+        setcolorder(res, setdiff(names(res), c("datetime", "value")))
+
+        res
     }))
 
     if (unit && nrow(dt)) {
@@ -406,14 +542,14 @@ get_nc_data <- function (x, lats, lons, years, unit = TRUE) {
     set(dt, NULL, c("variable", "description", "units"), list(var, var_long, units))
 
     # change column order
-    setcolorder(dt, c("datetime", "lat", "lon", "variable", "description", "units", "value"))
+    setcolorder(dt, c("index", "lon", "lat", "dist", "datetime", "variable", "description", "units", "value"))
 
     set(dt, NULL,
-        c("activity_drs", "experiment_id", "institution_id", "source_id", "member_id", "table_id"),
-        atts[J("NC_GLOBAL", c("activity_id", "experiment_id", "institution_id", "source_id", "variant_label", "frequency")),
+        c("activity_drs", "institution_id", "source_id", "experiment_id", "member_id", "table_id"),
+        atts[J("NC_GLOBAL", c("activity_id", "institution_id", "source_id", "experiment_id", "variant_label", "frequency")),
             on = c("variable", "attribute"), value]
     )
-    setcolorder(dt, c("activity_drs", "institution_id", "source_id", "experiment_id", "member_id", "table_id"))
+    setcolorder(dt, c("index", "activity_drs", "institution_id", "source_id", "experiment_id", "member_id", "table_id"))
 
     dt
 }
@@ -426,6 +562,9 @@ get_nc_data <- function (x, lats, lons, years, unit = TRUE) {
 #' [match_coord()] and extracts CMIP6 data using the coordinates and years of
 #' interest specified.
 #'
+#' `extract_data()` supports common calendars, including `365_day` and
+#' `360_day`, thanks to the [PCICt][PCICt::as.PCICt] package.
+#'
 #' `extract_data()` uses [future.apply][future.apply::future_lapply()]
 #' underneath. You can use your preferable future backend to
 #' speed up data extraction in parallel. By default, `extract_data()` uses
@@ -466,27 +605,29 @@ get_nc_data <- function (x, lats, lons, years, unit = TRUE) {
 #'
 #' @return An `epw_cmip6_data` object, which is basically a list of 3 elements:
 #'
-#' * `epw`: An [eplusr::Epw] object whose longitude and latitute are used to
+#' * `epw`: An [eplusr::Epw] object whose longitude and latitude are used to
 #'   extract CMIP6 data. It is the same object as created in [match_coord()]
-#' * `meta`: A list containing basic meta data of input EPW, including `city`,
-#'   `state_province`, `country`, `latitute` and `longitude`.
+#' * `meta`: A list containing basic metadata of input EPW, including `city`,
+#'   `state_province`, `country`, `latitude` and `longitude`.
 #' * `data`: An empty [data.table::data.table()] if `keep` is `FALSE` or a
-#'   [data.table::data.table()] of 12 columns if `keep` is `TRUE`:
+#'   [data.table::data.table()] of 14 columns if `keep` is `TRUE`:
 #'
-#'     | No.  | Column           | Type      | Description                                                          |
-#'     | ---: | -----            | -----     | -----                                                                |
-#'     | 1    | `activity_drs`   | Character | Activity DRS (Data Reference Syntax)                                 |
-#'     | 2    | `institution_id` | Character | Institution identifier                                               |
-#'     | 3    | `source_id`      | Character | Model identifier                                                     |
-#'     | 4    | `experiment_id`  | Character | Root experiment identifier                                           |
-#'     | 5    | `member_id`      | Character | A compound construction from `sub_experiment_id` and `variant_label` |
-#'     | 6    | `table_id`       | Character | Table identifier                                                     |
-#'     | 7    | `lat`            | Double    | Latitude of extracted location                                       |
-#'     | 8    | `lon`            | Double    | Latitude of extracted location                                       |
-#'     | 9    | `variable`       | Character | Variable identifier                                                  |
-#'     | 10   | `description`    | Character | Variable long name                                                   |
-#'     | 11   | `units`          | Character | Units of variable                                                    |
-#'     | 12   | `value`          | Double    | Start date and time of simulation                                    |
+#'     | No.  | Column           | Type      | Description                                                            |
+#'     | ---: | -----            | -----     | -----                                                                  |
+#'     | 1    | `activity_drs`   | Character | Activity DRS (Data Reference Syntax)                                   |
+#'     | 2    | `institution_id` | Character | Institution identifier                                                 |
+#'     | 3    | `source_id`      | Character | Model identifier                                                       |
+#'     | 4    | `experiment_id`  | Character | Root experiment identifier                                             |
+#'     | 5    | `member_id`      | Character | A compound construction from `sub_experiment_id` and `variant_label`   |
+#'     | 6    | `table_id`       | Character | Table identifier                                                       |
+#'     | 7    | `lon`            | Double    | Longitude of extracted location                                        |
+#'     | 8    | `lat`            | Double    | Latitude of extracted location                                         |
+#'     | 9    | `dist`           | Double    | The spherical distance in km between EPW location and grid coordinates |
+#'     | 10   | `datetime`       | POSIXct   | Datetime for the predicted value                                       |
+#'     | 11   | `variable`       | Character | Variable identifier                                                    |
+#'     | 12   | `description`    | Character | Variable long name                                                     |
+#'     | 13   | `units`          | Character | Units of variable                                                      |
+#'     | 14   | `value`          | Double    | The actual predicted value                                             |
 #'
 #' @importFrom checkmate assert_class
 #' @importFrom units set_units
@@ -526,7 +667,7 @@ extract_data <- function (coord, years = NULL, unit = FALSE, out_dir = NULL,
     }
 
     # initial progress bar
-    message("Start to extract CMIP6 data according to matched coordinates...")
+    verbose("Start to extract CMIP6 data according to matched coordinates...")
 
     data <- data.table()
 
@@ -534,20 +675,19 @@ extract_data <- function (coord, years = NULL, unit = FALSE, out_dir = NULL,
         co <- m_coord[[i]]
 
         if (!is.null(out_dir) && length(by_cols)) {
-            message("Extracting data for case '", names(m_coord)[[i]], "'...")
+            verbose("Extracting data for case '", names(m_coord)[[i]], "'...")
         }
         progressr::with_progress({
             p <- progressr::progressor(nrow(co))
-            ip <- 0L
 
             d <- rbindlist(
                 future.apply::future_Map(
-                    function (path, coord) {
-                        ip <<- ip + 1L
+                    function(ip, path, coord) {
                         p(message = sprintf("[%i/%i]", ip, nrow(co)))
-                        get_nc_data(path, lats = coord$lat, lons = coord$lon, years = years, unit = unit)
+                        d <- get_nc_data(path, coord, years = years, unit = unit)
+                        set(d, NULL, "index", NULL)
                     },
-                    co$file_path, co$coord
+                    seq.int(nrow(co)), co$file_path, co$coord
             ))
         })
 
@@ -652,7 +792,7 @@ get_nc_axes <- function (x) {
         on.exit(RNetCDF::close.nc(nc), add = TRUE)
     }
 
-    # wl wl get file info
+    # get file info
     inq <- RNetCDF::file.inq.nc(nc)
 
     vars <- rbindlist(lapply(seq_len(inq$nvars) - 1L, function (i) {
@@ -667,6 +807,7 @@ get_nc_axes <- function (x) {
 # }}}
 
 # get_nc_time {{{
+#' @importFrom PCICt as.PCICt
 get_nc_time <- function (x, range = FALSE) {
     if (inherits(x, "NetCDF")) {
         nc <- x
@@ -680,27 +821,53 @@ get_nc_time <- function (x, range = FALSE) {
     # get variable attributes
     atts <- get_nc_atts(nc)
 
+    # get calendar
+    cal <- tolower(tolower(atts[variable == "time" & attribute == "calendar", value[[1L]]]))
+    CALENDARS <- c("gregorian", "standard", "proleptic_gregorian", "noleap",
+        "365_day", "all_leap", "366_day", "360_day")
+    if (!cal %in% CALENDARS) {
+        stop(sprintf("Unsupported calendar type found: '%s'.", cal))
+    }
+
     # get time origin and units
-    ori <- atts[variable == "time" & attribute == "units", value[[1L]]]
+    time_spec <- atts[variable == "time" & attribute == "units", value[[1L]]]
+    time_units <- strsplit(time_spec, " ")[[1L]]
+    time_res <- tolower(time_units[1L])
 
-    # get first time value
-    time_start <- RNetCDF::var.get.nc(nc, "time", 1, 1)
+    if (length(time_units) >= 4L) {
+        time_ori_str <- paste(time_units[3L], time_units[4L:length(time_units)])
+    } else if (length(time_units) == 3L) {
+        time_ori_str <- time_units[3L]
+    } else {
+        stop(sprintf("Invalid time units found: '%s'.", time_spec))
+    }
+    # PCICt only handle standard format string
+    if (time_res %in% c("months", "month")) {
+        time_ori_str <- paste0(time_ori_str, "-01")
+        warning("Month time resolution found. Seconds in a month will be set to 86400 (seconds in day) * 30 (day). ",
+            "Time conversion may be not accurate.")
+    }
+    time_ori <- PCICt::as.PCICt.default(time_ori_str, cal = cal)
 
-    # get last time value
-    time_end <- RNetCDF::var.get.nc(nc, "time", n, 1)
+    secs <- switch(time_res,
+        second = 1, seconds = 1,
+        minute = 60, minutes = 60,
+        hour = 3600, hours = 3600,
+        day = 86400, days = 86400,
+        month = 86400 * 30, months = 86400 * 30,
+        stop(sprintf("Unsupported time resolution found: '%s'. Should be one of [%s]",
+            time_res, paste0("'", c("second(s)", "minute(s)", "hour(s)", "day(s)", "month(s)"), "'", collapse = ", ")
+        ))
+    )
 
     if (range) {
-        c(
-            RNetCDF::utcal.nc(ori, time_start, "c"),
-            RNetCDF::utcal.nc(ori, time_end, "c")
-        )
+        time_num_start <- RNetCDF::var.get.nc(nc, "time", 1, 1)
+        time_num_end <- RNetCDF::var.get.nc(nc, "time", n, 1)
+        as.POSIXct(time_ori + c(time_num_start, time_num_end) * secs)
     } else {
         # create time sequences
-        seq(
-            RNetCDF::utcal.nc(ori, time_start, "c"),
-            RNetCDF::utcal.nc(ori, time_end, "c"),
-            length.out = n
-        )
+        time_num <- RNetCDF::var.get.nc(nc, "time", 1, n)
+        as.POSIXct(time_ori + time_num * secs)
     }
 }
 # }}}
@@ -718,11 +885,11 @@ match_nc_time <- function (x, years = NULL) {
         i <- lapply(as.integer(years), function (x) which(y == x))
 
         j <- 1L
-        l <- list()
+        l <- vector("list", length(i))
         l[[1L]] <- i[[1L]]
         for (m in i[-1L]) {
             if (!length(m)) next
-            if (length(l[[j]]) && l[[j]][length(l[[j]])] + 1L == m[1L]) {
+            if (length(l[[j]]) && (l[[j]][length(l[[j]])] + 1L) == m[1L]) {
                 l[[j]] <- c(l[[j]], m)
             } else {
                 j <- j + 1L
@@ -735,6 +902,7 @@ match_nc_time <- function (x, years = NULL) {
 }
 # }}}
 
+# nocov start
 # rechunk_nc_dims {{{
 rechunk_nc_dims <- function (nc, out_file) {
     if (Sys.which("nccopy") == "") return()
@@ -783,3 +951,4 @@ reorganize_nc_dims <- function (nc) {
     re
 }
 # }}}
+# nocov end
